# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#=============================================================================== GraphQL handlers =========================================================================================================
from .fx_types import Effect, FX
from .._ops import *
from .http import send_response, permit_cors, middleware, middleware_worker, fallback_not_found
from ariadne import graphql_sync
import os
import json

from ..logger import log


def graphql_start_server_handler(eff: Effect):
    """Example
    {
        "type": FX.GraphQL.StartServer,
        "port": 5001, (default=5000) 
        "schema_root": z_schema,
        "path": "/gql", (default="/gql")
        "playground_path": "/", (default=None)
        "open_browser": False (default=False)
    } | run
    """


    from ...gql.generate_gql_api import make_api
    schema = make_api(eff['schema_root'])
    port = eff.get("port", 5000)

    gql_path = eff.get("path", "/gql")
    playground_path = eff.get("playground_path", None)
    open_browser = eff.get("open_browser", False)

    if not gql_path.startswith("/"):
        raise Exception("GQL path must start with /")
    if playground_path is not None and not playground_path.startswith("/"):
        raise Exception("Playground path must start with /")

    def resolve_gql_query(req, port=port, playground_path=playground_path, gql_path=gql_path):
        import copy
        req = copy.deepcopy(req)
        if playground_path is not None and req["path"] == playground_path:
            from ... import gql
            import os
            path = os.path.join(*gql.__path__, "playground.html")
            with open(path) as file:
                body = file.read()
            body = body.replace("ZEF_PORT", str(port))
            body = body.replace("ZEF_GQL_PATH", gql_path)
            req["response_body"] = body
        elif req["path"] == gql_path:
            success, result = graphql_sync(schema, json.loads(req["request_body"]))
            req["response_body"] = json.dumps(result)
        return req


    logging = eff.get("logging", False)
    http_r = {
        **eff,     # Verify that this is ok after change: the dict was mutated previously (using pop). There may be more key-value pairs in there now.
        'type': FX.HTTP.StartServer,
        'port': port,
        # 'pipe_into': map[resolve_gql_query] | subscribe[run]
        # This doesn't work with rxops atm... need to do it all in a custom map[] function
        # 'pipe_into': middleware[[permit_cors, resolve_gql_query, send_response]]
        'pipe_into': map[middleware_worker[permit_cors, resolve_gql_query, fallback_not_found, send_response]] | subscribe[run],
        'logging': logging,
    } | run

    log.info(f"Started GQL server at http://localhost:{port}{gql_path}")
    if playground_path is not None:
        playground_url = f"http://localhost:{port}{playground_path}"
        log.info(f"Playground available at {playground_url}")
        if open_browser:
            import webbrowser
            webbrowser.open(playground_url)

    return http_r
    
    
def graphql_stop_server_handler(eff: dict):    
    return run({**eff, 'type': FX.HTTP.StopServer})

    
def graphql_start_playground_handler(eff: dict):
    """Example
    {
        "type": FX.GraphQL.StartPlayground,
        "schema_root": z_schema,
        "port": 5001, (default=5000)
        "path": "/gql", (default="/gql")
        "playground_path": "/", (default="/")
        "open_browser": True (default=True)
    } | run
    """   
    schema = eff['schema_root']
    port = eff.get("port", 5000)

    gql_path = eff.get("path", "/gql")
    playground_path = eff.get("playground_path", "/") # It is required!
    open_browser = eff.get("open_browser", True)

    if not gql_path.startswith("/"):
        raise Exception("GQL path must start with /")
    if not playground_path.startswith("/"):
        raise Exception("Playground path must start with /")

    http_r = {
        "type": FX.GraphQL.StartServer,
        "port": port,
        "schema_root": schema,
        "path": gql_path,
        "playground_path": playground_path,
        "open_browser": open_browser,
    } | run
    
    return http_r
    
    
def graphql_stop_playground_handler(eff: dict):    
    return run({**eff, 'type': FX.HTTP.StopServer})
    

def graphql_generate_schema_string_handler(eff: dict):   
    """
    Given a ZefRef to a ET.GQL_Schema node, this effect generates a
    GraphQL string representation of the schema.

    Usage:

    Effect({
        type: FX.GraphQL.GenerateSchemaString,
        schema_root: zr_schema_root,
    })
    """
    from zef.gql.schema_graph_to_file import generate_api_schema_string
    schema_root = eff["schema_root"] 
    return generate_api_schema_string(schema_root)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

class MyTestCase(unittest.TestCase):
    def test_partial_graphs_same(self):
        g = Graph()

        g_clones = []
        for i in range(10):
            g_clones += [zef.pyzef.internals.create_partial_graph(g.graph_data, g.graph_data.write_head)]
            [ET.Machine]*10 | g | run

        # Add in a couple of manual edge cases
        g_clones += [zef.pyzef.internals.create_partial_graph(g.graph_data, g.graph_data.write_head)]
        z = ET.Machine | g | run
        (z, RT.Something, 1) | g | run
        g_clones += [zef.pyzef.internals.create_partial_graph(g.graph_data, g.graph_data.write_head)]
        [(z, RT.Something, z), (z, RT.Something2, z)] | g | run
        g_clones += [zef.pyzef.internals.create_partial_graph(g.graph_data, g.graph_data.write_head)]
        z | terminate | g | run
        g_clones += [g]

        for i,tx in enumerate(g | all[TX] | skip[1] | collect):
            g_clone_before_tx = g_clones[i]

            before_heads = zef.internals.create_update_heads(g_clone_before_tx.graph_data)
            before_payload = zef.internals.create_update_payload(g_clone_before_tx.graph_data, before_heads, "")
            for j in range(i+1,len(g_clones)):
                g_partial = zef.pyzef.internals.create_partial_graph(g_clones[j].graph_data, g_clone_before_tx.graph_data.write_head)
                self.assertEqual(g_clone_before_tx.graph_data.hash(), g_partial.graph_data.hash())

                after_payload = zef.internals.create_update_payload(g_partial.graph_data, before_heads, "")
                self.assertEqual(before_payload, after_payload)

    def test_partial_during_transaction(self):
        g = Graph()

        a,b,c = (ET.Machine, RT.Something, 5) | g | run
        hash_before = g.hash()
        head_before = g.graph_data.read_head
        with Transaction(g):
            a | terminate | g | run
            [ET.Machine]*10 | g | run
            a,b,c = (ET.Machine, RT.Something, 5) | g | run
            g_clone_in_tx = zef.pyzef.internals.create_partial_graph(g.graph_data, head_before)
            hash_in_tx = zef.pyzef.internals.partial_hash(g, head_before)
        g_clone_after = zef.pyzef.internals.create_partial_graph(g.graph_data, head_before)
        hash_after = zef.pyzef.internals.partial_hash(g, head_before)
            
        self.assertEqual(g_clone_after.graph_data.hash(), hash_before)
        self.assertEqual(g_clone_in_tx.graph_data.hash(), hash_before)
        self.assertEqual(hash_after, hash_before)
        self.assertEqual(hash_in_tx, hash_before)

    def test_partial_from_merges(self):
        g = Graph()
        g2 = Graph()

        a,b,c = (ET.Machine, RT.Something, 5) | g | run
        a2,b2,c2 = (ET.Machine, RT.Something, 5) | g2 | run
        
        hash_before = g.hash()
        head_before = g.graph_data.read_head
        with Transaction(g):
            a | terminate | g | run
            [ET.Machine]*10 | g | run
            a3,b3,c3 = [a2,b2,c2] | g | run
        with Transaction(g):
            terminate(a3) | g | run
        g_clone_after = zef.pyzef.internals.create_partial_graph(g.graph_data, head_before)
        hash_after = zef.pyzef.internals.partial_hash(g, head_before)
            
        self.assertEqual(g_clone_after.graph_data.hash(), hash_before)
        self.assertEqual(hash_after, hash_before)

    def test_abort_transaction(self):
        g = Graph()

        z = ET.Machine | g | run

        before_hash = g.hash()

        with Transaction(g):
            z2 = ET.Machine | g | run
            zef.pyzef.internals.AbortTransaction(g)
        self.assertEqual(before_hash, g.hash())

        with Transaction(g):
            (z, RT.Something, 5) | g | run
            zef.pyzef.internals.AbortTransaction(g)
        self.assertEqual(before_hash, g.hash())

        with Transaction(g):
            z | terminate | g | run
            zef.pyzef.internals.AbortTransaction(g)
        self.assertEqual(before_hash, g.hash())

        dz = delegate_of(z)
        (dz, RT.Meta, 1) | g | run
        before_hash = g.hash()

        with Transaction(g):
            (dz, RT.Something, 5) | g | run
            zef.pyzef.internals.AbortTransaction(g)
        self.assertEqual(before_hash, g.hash())

        z,_,ae = (z, RT.Something, 1) | g | run
        before_hash = g.hash()

        with Transaction(g):
            ae | assign[2] | g | run
            zef.pyzef.internals.AbortTransaction(g)
        self.assertEqual(before_hash, g.hash())

    def test_empty_transaction(self):
        g = Graph()

        num_txs = len(g | all[TX] | collect)
        before_read_head = g.graph_data.read_head

        with Transaction(g, True, True):
            pass

        self.assertEqual(num_txs, len(g | all[TX] | collect))
        self.assertEqual(before_read_head, g.graph_data.read_head)

        with Transaction(g, True, False):
            pass

        self.assertEqual(True, num_txs < len(g | all[TX] | collect))
        self.assertEqual(True, before_read_head < g.graph_data.read_head)

        

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
In Zef the types of Zef Values are values themselves: ValueType_.
These are distinct from the host language's implementation type
of the value instance.

In contrast to most programming languages's type system, the 
Zef Type system is based on set theory and has a stronger focus 
on succinctly being able to express complex domains and associative
ontolgies.

ValueTypes fall into the category of Zef Values that can absorb other
Zef Values, which is useful in constructing query expressions (as native
data structures) and graph commands. 
It also allows them to be flexibly used by the user to build custom DSLs.

---------------------- internal structure ---------------------------
It also has "absorbed": a tuple of other values


----------- Nesting Order ----------------
TODO: Do we want to keep this? Probably not.


What is "Union"? A basic value type.
Union[Int, Union[Float, String]]

Union[Int,Float]
Union[Int,Float,String]

This can be distinguished from Tuples as a type
Union[(Int,Float,String),]


----------- Constructor Overloading ----------------
Int(3.4) can have overloaded behavior. The user 
provides a function.

------------ Square Bracket Overloading --------
Should Union be of type ZefType? Or just return a ZefType?


Intersection
Union
Is

There is also a Union ValueType_, which is returned.
This custom structure may rearrange and parse terms.


---------- User Defined Types ----------
needed?
"""
from ..error import Error
from .._core import *
from ..internals import *


# For certain value types, the user may want to call e.g. Float(4.5).
# This looks up the associated function 
_value_type_constructor_funcs = {}
_value_type_get_item_funcs = {}

class ValueType_:
    """ 
    Zef ValueTypes are Values themselves.
    """
    def __init__(self, type_name: str, absorbed=(), constructor_func=None, get_item_func=None, pass_self=False):
            self.d = {
                'type_name': type_name,
                'absorbed': absorbed,
                'alias': None,
            }
            if constructor_func is not None:
                _value_type_constructor_funcs[type_name] = constructor_func
            if get_item_func is not None:
                _value_type_get_item_funcs[type_name] = get_item_func

            self.pass_self = pass_self
            self.allowed_types = (ValueType_, EntityTypeStruct, RelationTypeStruct, AttributeEntityTypeStruct, BlobTypeStruct, BlobType, AttributeEntityType, EntityType, RelationType)

    def __repr__(self):
        if self.d['alias'] != None:
            return self.d['alias']
        return self.d['type_name'] + (
            '' if self.d['absorbed'] == () 
            else ''.join([ f"[{repr(el)}]" for el in self.d['absorbed'] ])
        )


    def __call__(self, *args, **kwargs):
        try:
            f = _value_type_constructor_funcs[self.d["type_name"]]
        except KeyError:
            return Error(f'{self.d["type_name"]}(...) was called, but no constructor function was registered for this type')
        if self.pass_self: return f(self, *args, **kwargs)
        return f(*args, **kwargs)


    def __getitem__(self, x):
        try:
            f = _value_type_get_item_funcs[self.d["type_name"]]
            if self.pass_self: return f(self, x)
            return f(x)
        except KeyError:
            return ValueType_(self.d["type_name"], absorbed=(*self.d['absorbed'], x))


    def __eq__(self, other):
        if not isinstance(other, ValueType_): return False
        return self.d['type_name'] == other.d['type_name'] and self.d['absorbed'] == other.d['absorbed']


    def __hash__(self):
        return hash(self.d['type_name']) ^ hash(self.d['absorbed'])


    def __or__(self, other):
        from ..op_structs import ZefOp
        if isinstance(other, self.allowed_types):
            return simplify_value_type(ValueType_(type_name='Union', absorbed=(self, other,)))
        elif isinstance(other, ZefOp):
            return other.__ror__(self)
        else:
            raise Exception(f'"ValueType_`s "|" called with unsupported type {type(other)}')

    def __ror__(self, other):
        # Is | commutative? Going to assume it isn't
        if isinstance(other, self.allowed_types):
            return simplify_value_type(ValueType_(type_name='Union', absorbed=(other, self,)))
        else:
            raise Exception(f'"ValueType_`s right-side "|" called with unsupported type {type(other)}')
    
    def __and__(self, other):
        if isinstance(other, self.allowed_types):
            return simplify_value_type(ValueType_(type_name='Intersection', absorbed=(self, other,)))
        else:
            raise Exception(f'"ValueType_`s "&" called with unsupported type {type(other)}')

    def __rand__(self, other):
        # Is & commutative? Going to assume it isn't
        if isinstance(other, self.allowed_types):
            return simplify_value_type(ValueType_(type_name='Intersection', absorbed=(other, self,)))
        else:
            raise Exception(f'"ValueType_`s right-side "&" called with unsupported type {type(other)}')

    def __invert__(self):
        return ValueType_(type_name='Complement', absorbed=(self,))


    def __contains__(self, x):
        """
        Allows checking membership in form
        >>> 4 in Int
        """
        from ..op_implementations.implementation_typing_functions import is_a_implementation
        return is_a_implementation(x, self)
    

def make_distinct(v):
    """
    utility function to replace the 'distinct'
    zefop and preserve order.
    """
    seen = set()
    for el in v:
        if isinstance(el, set):
            # we can't import the following at the top of file or even outside this if
            # block: this function executes upon import of zef and leads to 
            # circular dependencies. We do really want to use the value_hash zefop though.
            from ..op_implementations.implementation_typing_functions import value_hash
            h = value_hash(el)
        else:
            h = el
        if h not in seen:
            yield el
            seen.add(h)


def simplify_value_type(x):
    """
    Only simplifies nested Union and Intersection.
    It does not change any absorbed values, but
    only rearranges on the outer level of Zef Values.

    Union and Intersection behave analogous to + and *.

    I[U[A, B], C] = I[U[A,C], U[B,C]]              # True
    I[U[Ev, Od], Pos] = I[U[Ev,Pos], U[Odd,Pos]]   # True
    (A+B)*C = A*C + B*C
    
    U[I[Ev, Od], Pos] = U[I[Ev,Pos], I[Odd,Pos]]   # This is False!!!!


    """
    def is_a_union(y):
        try:
            return y.d['type_name'] == 'Union'
        except:
            return False
    def is_a_intersection(y):
        try:
            return y.d['type_name'] == 'Intersection'
        except:
            return False

    if is_a_union(x):
        # flatten out unions: Union[Union[A][B]][C]  == Union[A][B][C]
        old_abs = x.d['absorbed']
        absorbed = tuple((el.d['absorbed'] if is_a_union(el) else (el,) for el in old_abs))  # flatten this out
        return ValueType_(type_name='Union', absorbed=tuple(make_distinct((a2 for a1 in absorbed for a2 in a1))))
    elif is_a_intersection(x):
        # flatten out Intersections: Intersection[Intersection[A][B]][C]  == Intersection[A][B][C]
        old_abs = x.d['absorbed']
        absorbed = tuple((el.d['absorbed'] if is_a_intersection(el) else (el,) for el in old_abs))  # flatten this out
        return ValueType_(type_name='Intersection', absorbed=tuple(make_distinct((a2 for a1 in absorbed for a2 in a1))))
    else:
        return x



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This is an expression type / value that most users do not have to interact with directly.

There are two main use cases for it at the moment:

A)  It allows for a more readable syntax when writing queries.
    The syntax is similar to that in Pandas and R's Deplyr
    .
    OldPerson = ET.Person & (z.Age > 60)
    
    g = Graph()
    [{
        ET.Movie: {
            RT.Title: 'Pulp Fiction',
            RT.Director: { ET.Person: { RT.FirstName: 'Quentin', RT.Age: 60 } },
        },    
    },
    {ET.Person: { RT.FirstName: 'Fred', RT.Age: 30 }},
    ] | g | run
    g | now | all[OldPerson] | c

    Why can we not just use SymbolicExpressions for this? We want `(z.Age > 60)` to 
    immediate evaluate to a ValueType. If this were not the case, this would break
    the type system / add massive complications. Therefore this new type is created.

B)  We could also use this type as variables in queries, to which values are 
    bound within some context.
"""

from .VT import  Is
from ._ops import equals, greater_than, less_than, greater_than_or_equal, less_than_or_equal, collect, F, alias

class ZField_:
    """
    make this a special type: we want to use it in expressions like
    (z.age > 10)   # and this should evaluate to a ValueType.

    This is also used in structural pattern matching.
    """
    def __init__(self, name):
        self.name = name
    
    def __repr__(self):
        return f'z.{self.name}'

    def __eq__(self, other):
        return (Is[F.__getattr__(self.name) | equals[other] ] 
        | alias[f"(z.{self.name} == {other})"] 
        | collect
        )

    def __gt__(self, other):
        return (Is[F.__getattr__(self.name) | greater_than[other] ] 
        | alias[f"(z.{self.name} > {other})"] 
        | collect
        )

    def __ge__(self, other):
        return (Is[F.__getattr__(self.name) | greater_than_or_equal[other] ] 
        | alias[f"(z.{self.name} >= {other})"] 
        | collect
        )

    def __lt__(self, other):
        return (Is[F.__getattr__(self.name) | less_than[other] ] 
        | alias[f"(z.{self.name} < {other})"] 
        | collect
        )

    def __le__(self, other):
        return (Is[F.__getattr__(self.name) | less_than_or_equal[other] ] 
        | alias[f"(z.{self.name} <= {other})"] 
        | collect
        )





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = [
    "take_transactor_role",
    "release_transactor_role",
    "have_transactor_role",
]

from ..core import *
from ..core._ops import *

from ..core.op_implementations.dispatch_dictionary import _op_to_functions

def take_transactor_role_implementation(g):
    return {"type": FX.Graph.TakeTransactorRole, "graph": g}
    
def take_transactor_role_tp(x):
    return VT.Any

def release_transactor_role_implementation(g):
    return {"type": FX.Graph.ReleaseTransactorRole, "graph": g}

def release_transactor_role_tp(x):
    return VT.Any

def have_transactor_role_implementation(g):
    return g.graph_data.is_primary_instance
def have_transactor_role_tp(x):
    return VT.Bool

_op_to_functions[RT.TakeTransactorRole] = (take_transactor_role_implementation, take_transactor_role_tp)
_op_to_functions[RT.ReleaseTransactorRole] = (release_transactor_role_implementation, release_transactor_role_tp)
_op_to_functions[RT.HaveTransactorRole] = (have_transactor_role_implementation, have_transactor_role_tp)

take_transactor_role = make_zefop(RT.TakeTransactorRole)
release_transactor_role = make_zefop(RT.ReleaseTransactorRole)
have_transactor_role = make_zefop(RT.HaveTransactorRole)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


#---------------------------------------------------------------------------------------

# allow value based comparisons
class FXElement():
    def __init__(self, args):
        """the data is stored as an array, typically of length 2"""
        self.d = args
        
    def __repr__(self):
        return "FX." + '.'.join(self.d)

    # This is a check to make sure we align the strings with the variable names.
    # We could also use this to assign the strings in the first place
    def __set_name__(self, owner, name):
        assert self.d[0] == owner._name
        assert self.d[1] == name


    def __call__(self, **kwargs):
        """
        Allow using this type as a constructor function,
        which automatically injects its own type. Other
        key-value pairs are passed in as keyword arguments.
        e.g.
        >>> FX.HTTP.Request(url='zefhub.io')     # returns an Effect

        TODO: if the effect module provides validation functions that
        can be run, execute this before returning. If it fails, return an
        error.
        """
        return {
            **kwargs,
            'type': self,
        }



class _HTTP_Class():
    _name = "HTTP"
    StartServer = FXElement(('HTTP', 'StartServer'))
    StopServer = FXElement(('HTTP', 'StopServer'))
    SendResponse = FXElement(('HTTP', 'SendResponse'))
    
    Request = FXElement(('HTTP', 'Request'))


class _Websocket_Class():
    _name = "Websocket"
    ConnectToServer = FXElement(('Websocket', 'ConnectToServer'))
    StartServer = FXElement(('Websocket', 'StartServer'))
    StopServer = FXElement(('Websocket', 'StopServer'))
    SendMessage = FXElement(('Websocket', 'SendMessage'))
    CloseConnections = FXElement(('Websocket', 'CloseConnections'))

class _Subprocess_Class():
    _name = "Subprocess"
    Start = FXElement(('Subprocess', 'Start'))
    Stop = FXElement(('Subprocess', 'Stop'))

class _ZefUI_Class():
    _name = "ZefUI"
    StartApp = FXElement(('ZefUI', 'StartApp'))
    StartServer = FXElement(('ZefUI', 'StartServer'))
    StopServer = FXElement(('ZefUI', 'StopServer'))


class _GraphQL_Class():
    _name = "GraphQL"
    StartServer = FXElement(('GraphQL', 'StartServer'))
    StopServer = FXElement(('GraphQL', 'StopServer'))
    StartPlayground = FXElement(('GraphQL', 'StartPlayground'))
    StopPlayground = FXElement(('GraphQL', 'StopPlayground'))
    GenerateSchemaString = FXElement(('GraphQL', 'GenerateSchemaString'))


class _LocalFile_Class():
    _name = "LocalFile"
    Read = FXElement(('LocalFile', 'Read'))
    ReadBinary = FXElement(('LocalFile', 'ReadBinary'))
    Load = FXElement(('LocalFile', 'Load'))
    Write = FXElement(('LocalFile', 'Write'))
    Save = FXElement(('LocalFile', 'Save'))
    SystemOpenWith = FXElement(('LocalFile', 'SystemOpenWith'))
    MonitorPath     = FXElement(('LocalFile', 'MonitorPath'))
    

class _S3_Class():
    _name = "S3"
    Read = FXElement(('S3', 'Read'))
    Write = FXElement(('S3', 'Write'))
    
    
class _FX_Class():
    _name = "FX"
    AddEffectHandler = FXElement(('FX', 'AddEffectHandler'))

class _Graph_Class():
    _name = "Graph"
    Tag = FXElement(('Graph', 'Tag'))
    Sync = FXElement(('Graph', 'Sync'))
    TakeTransactorRole = FXElement(('Graph', 'TakeTransactorRole'))
    ReleaseTransactorRole = FXElement(('Graph', 'ReleaseTransactorRole'))
    Transact = FXElement(('Graph', 'Transact'))
    Load = FXElement(('Graph', 'Load'))
    # Or does it make more sense to write FX.Tag.Graph? Tagging of RAEs definitely belongs into a GraphDelta / graph tx though

class _Stream_Class():
    _name = "Stream"
    CreatePushableStream = FXElement(('Stream', 'CreatePushableStream'))
    Push = FXElement(('Stream', 'Push'))
    Complete = FXElement(('Stream', 'Complete'))
    

class _Privileges_Class():
    _name = "Privileges"
    GrantView = FXElement(('Privileges', 'GrantView'))
    GrantAppend = FXElement(('Privileges', 'GrantAppend'))
    GrantHost = FXElement(('Privileges', 'GrantHost'))
    GrantDiscover = FXElement(('Privileges', 'GrantDiscover'))
    GrantModifyRights = FXElement(('Privileges', 'GrantModifyRights'))
    RevokeView = FXElement(('Privileges', 'RevokeView'))
    RevokeAppend = FXElement(('Privileges', 'RevokeAppend'))
    RevokeHost = FXElement(('Privileges', 'RevokeHost'))
    RevokeDiscover = FXElement(('Privileges', 'RevokeDiscover'))
    RevokeModifyRights = FXElement(('Privileges', 'RevokeModifyRights'))

class _ZefHub_Class():
    _name = "ZefHub"
    Login = FXElement(('ZefHub', 'Login'))
    Logout = FXElement(('ZefHub', 'Logout'))


# using the following, some IDEs don't pick up the class members
# class _Placeholder_Class():
#     def __init__(self, group_name:str, v):
#         self.A = FXElement(('Placeholder', 'A'))
#         setattr(self, 'B', FXElement(('Placeholder', 'B')))


class _Clipboard_Class():
    _name = "Clipboard"
    CopyTo = FXElement(('Clipboard', 'CopyTo'))
    CopyFrom = FXElement(('Clipboard', 'CopyFrom'))     # prefer this over "paste", which is imperative to 'paste' it somewhere, i.e. plop. We can also just get a value out, i.e. this is a data source in the imperative shell

class _RAE_Class():
    _name = "RAE"
    Tag = FXElement(('RAE', 'Tag'))



class _FX_Class:
    HTTP = _HTTP_Class()
    Websocket = _Websocket_Class()
    Subprocess = _Subprocess_Class()
    GraphQL = _GraphQL_Class()
    ZefUI = _ZefUI_Class()
    LocalFile = _LocalFile_Class()
    S3 = _S3_Class()
    Graph = _Graph_Class()
    FX = _FX_Class()
    Stream = _Stream_Class()
    Privileges = _Privileges_Class()
    Clipboard = _Clipboard_Class()
    ZefHub = _ZefHub_Class()
    RAE = _RAE_Class()


FX = _FX_Class()

_group_types = [ _Clipboard_Class,_FX_Class,_GraphQL_Class,_Graph_Class,_HTTP_Class,_LocalFile_Class,_Privileges_Class,_S3_Class,_Stream_Class,_Websocket_Class,_ZefHub_Class,_ZefUI_Class]
#---------------------------------------------------------------------------------------
    

# just a function, since Effect will just become a value.
# same syntax as a regular constructor and equivalent to to_effect.
# The latter is the more usual option to use in the context of piping
def Effect(*args, **kwargs) -> dict:
    """
    Effects can be constructed via keyword arguments
    or by passing in a dictionary.
    
    ---- Examples ----
    FX.Subprocess.Start()       # the call operator acts like a constructor. The type is inserted automatically.
    """
    if args!=():
        assert len(kwargs) == 0
        assert len(args) == 1
        assert isinstance(args[0], dict)
        return args[0]

    else:
        return kwargs




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# This file contains tests for things that were bugs at one point. Does not need
# to be organised at all

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
import os
from zef import *
from zef.ops import *
import zef


class MyTestCase(unittest.TestCase):

    def test_terminate_relation(self):
        g = Graph()

        r = [
            ET.Person["joe"],
            (Z["joe"], RT.FirstName["rel"], "name"),
        ] | transact[g] | run

        self.assertEqual(1, g | now | all[ET.Person] | length | collect)
        self.assertEqual(1, g | now | all[RT.FirstName] | length | collect)

        r["rel"] | terminate | g | run
        self.assertEqual(1, g | now | all[ET.Person] | length | collect)
        self.assertEqual(0, g | now | all[RT.FirstName] | length | collect)

    def test_terminate_relation_merge(self):
        g = Graph()

        r = [
            ET.Person["joe"],
            (Z["joe"], RT.FirstName["rel"], "name"),
        ] | transact[g] | run

        self.assertEqual(1, g | now | all[ET.Person] | length | collect)
        self.assertEqual(1, g | now | all[RT.FirstName] | length | collect)

        g2 = Graph()
        r2 = [
            r["rel"] | terminate["rel from g2"],
        ] | transact[g2] | run
        self.assertEqual(1, g | now | all[ET.Person] | length | collect)
        self.assertEqual(1, g | now | all[RT.FirstName] | length | collect)
        self.assertEqual(0, g2 | now | all[ET.Person] | length | collect)
        self.assertEqual(0, g2 | now | all[RT.FirstName] | length | collect)
        self.assertEqual(None, r2["rel from g2"])

    def test_no_duplicate_internal_name(self):
        g = Graph()
        with self.assertRaises(Exception):
            r = [
                ET.Danny["a"],
                ET.Machine["a"],
                (Z["a"], RT.Something, "name")
            ] | transact[g] | run
        

if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import Effect
from ..error import Error

def graph_tag_handler(eff: Effect):
    try:
        from ...pyzef.main import tag
        tag(eff["graph"], eff["tag"], adding=eff["adding"], force=eff["force"])
        return {}
    except Exception as e:
        return Error(f'executing FX.Graph.Tag for effect {eff}:\n{repr(e)}')




def rae_tag_handler(eff: Effect):
    try:
        from ...pyzef.main import tag
        if not eff["adding"]:
            return Error("Untagging a RAE is not supported (yet).")
        tag(eff["rae"], eff["tag"], force_if_name_tags_other_rel_ent=eff["force"])
        return {}
    except Exception as e:
        return Error(f'executing FX.RAE.Tag for effect {eff}:\n{repr(e)}')




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *
import zef

class MyTestCase(unittest.TestCase):

    def test_value_node_reuse(self):
        g = Graph()

        z = zef.pyzef.main.instantiate_value_node(5, g)
        z2 = zef.pyzef.main.instantiate_value_node(5, g)

        self.assertEqual(z, z2)

        from zef.pyzef.zefops import SerializedValue
        s = SerializedValue("plain string", "asdf")
        z = zef.pyzef.main.instantiate_value_node(s, g)
        z2 = zef.pyzef.main.instantiate_value_node(s, g)

        self.assertEqual(z, z2)

    def test_value_node_graph_delta(self):
        g = Graph()

        z = Val(5) | g | run
        z2 = Val(5) | g | run

        self.assertEqual(z, z2)
        self.assertEqual(value(z), 5)

        z = Val({'some_dict': 5}) | g | run
        z2 = Val({'some_dict': 5}) | g | run

        self.assertEqual(z, z2)
        self.assertEqual(value(z), {'some_dict': 5})

        a,b,c = (ET.Machine, RT.Kind, Val("operating")) | g | run
        self.assertEqual(value(c), "operating")

        d,e,f = (ET.Machine, RT.Kind, Val("operating")) | g | run
        self.assertEqual(now(c), now(f))

        self.assertEqual(d | Out[RT.Kind] | Ins[RT.Kind] | func[set] | collect,
                         {now(a), now(d)})

        self.assertEqual(g | now | all | length | collect, 7)

    def test_type_nodes(self):
        g = Graph()

        ae = AET.Type | g | run
        ae | assign[AET.Int] | g | run

        self.assertEqual(value(ae | now), AET.Int)

    def test_logic_type(self):
        g = Graph()

        even_type = Int & Is[modulo[2] | equals[0]]

        ae = AET[even_type] | g | run

        self.assertEqual(AET(ae).complex_value.deserialize(), even_type)

        ae | assign[2] | g | run
        with self.assertRaises(Exception):
            ae | assign[3] | g | run
        ae | assign[4.0] | g | run

        self.assertEqual(type(value(ae | now)), float)
        self.assertEqual(type(value(ae | now | time_travel[-1])), int)

    def test_logic_type_hacked(self):
        import json
        from zef.pyzef.zefops import SerializedValue

        g = Graph()

        even_type = Int & Is[modulo[2] | equals[0]]
        s_typ = SerializedValue.serialize(even_type)

        ae = instantiate(AttributeEntityType(s_typ), g)

        ae | assign[2] | g | run
        with self.assertRaises(Exception):
            ae | assign[3] | g | run
        ae | assign[4.0] | g | run

        self.assertEqual(type(value(ae | now)), float)
        self.assertEqual(type(value(ae | now | time_travel[-1])), int)

    def test_assign_value_node(self):
        g = Graph()

        ae = AET.Int | g | run

        vn = zef.pyzef.main.instantiate_value_node(2, g)
        vn2 = zef.pyzef.main.instantiate_value_node(3.0, g)

        ae | assign[vn] | g | run
        ae | assign[vn2] | g | run

        ae | assign[Val(5)] | g | run

        self.assertEqual(type(value(ae | now)), int)
        self.assertEqual(type(value(ae | now | time_travel[-1])), int)
        self.assertEqual(type(value(ae | now | time_travel[-2])), int)

        # There shouldn't be an extra value node after assigning the same value.
        g = Graph()
        num_before = g | all | filter[is_a[BT.VALUE_NODE]] | length | collect
        vn = Val({"something": 42}) | g | run
        ae = AET.Any | g | run
        ae | assign[Val({"something": 42})] | g | run
        num_after = g | all | filter[is_a[BT.VALUE_NODE]] | length | collect
        self.assertEqual(num_after, num_before + 2)

        # There shouldn't be an extra value node if the value is used as a value
        # and a value type
        g = Graph()
        num_before = g | all | filter[is_a[BT.VALUE_NODE]] | length | collect
        vn = Val(VT.Graph) | g | run
        ae = AET.Any | g | run
        ae2 = AET[VT.Graph] | g | run
        ae | assign[Val(VT.Graph)] | g | run
        num_after = g | all | filter[is_a[BT.VALUE_NODE]] | length | collect
        self.assertEqual(num_after - num_before, 2)

if __name__ == '__main__':
    unittest.main()


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .fx_types import FX, Effect

from .state import _state


# from .http import (
#     http_start_server_handler,
#     http_stop_server_handler,
#     http_send_response_handler,
# )
from .http import (
    http_start_server_handler,
    http_stop_server_handler,
    http_send_response_handler,
    http_send_request_handler
)


from .websocket import (
    websocket_connect_to_server_handler,
    websocket_start_server_handler, 
    websocket_stop_server_handler, 
    websocket_send_message_handler, 
    websocket_close_connections_handler,
    )


from .graphql import (
    graphql_start_server_handler,
    graphql_stop_server_handler,
    graphql_start_playground_handler,
    graphql_stop_playground_handler,
    graphql_generate_schema_string_handler,
)


from .stream import (
    stream_create_pushable_stream_handler,
    stream_push_handler,
    stream_complete_handler
)

from .privileges import (
    privileges_grantview_handler,
    privileges_grantappend_handler,
    privileges_granthost_handler,
    privileges_grantdiscover_handler,
    privileges_grantmodifyrights_handler,
    privileges_revokeview_handler,
    privileges_revokeappend_handler,
    privileges_revokehost_handler,
    privileges_revokediscover_handler,
    privileges_revokemodifyrights_handler,
)

from .zefhub import (
    zefhub_login_handler,
    zefhub_logout_handler,
)

from .tag import (
    graph_tag_handler,
    rae_tag_handler,
)

from .sync import (
    graph_sync_handler,
)

from .graph import (
    graph_take_transactor_role_handler,
    graph_release_transactor_role_handler,
    graph_transaction_handler,
    graph_load_handler,
)


from .clipboard import (
    clipboard_copy_to_handler,
    clipboard_copy_from_handler,
)

from .local_file import (
    read_localfile_handler,
    readbinary_localfile_handler,
    write_localfile_handler,
    save_localfile_handler,
    load_localfile_handler,
    system_open_with_handler,
    monitor_path_handler,
)

# note the ".d" to access the tuple of Strings!
_effect_handlers = {
    FX.HTTP.StartServer.d: http_start_server_handler,
    FX.HTTP.StopServer.d: http_stop_server_handler,
    FX.HTTP.SendResponse.d: http_send_response_handler,
    FX.HTTP.Request.d: http_send_request_handler,
    
    FX.Websocket.ConnectToServer.d: websocket_connect_to_server_handler,
    FX.Websocket.StartServer.d: websocket_start_server_handler,
    FX.Websocket.StopServer.d: websocket_stop_server_handler,
    FX.Websocket.SendMessage.d: websocket_send_message_handler,
    FX.Websocket.CloseConnections.d: websocket_close_connections_handler,
    
    FX.GraphQL.StartServer.d: graphql_start_server_handler,
    FX.GraphQL.StopServer.d: graphql_stop_server_handler,
    FX.GraphQL.StartPlayground.d: graphql_start_playground_handler,
    FX.GraphQL.StopPlayground.d: graphql_stop_playground_handler,
    FX.GraphQL.GenerateSchemaString.d: graphql_generate_schema_string_handler,
    
    FX.Stream.CreatePushableStream.d: stream_create_pushable_stream_handler,
    FX.Stream.Push.d: stream_push_handler,
    FX.Stream.Complete.d: stream_complete_handler,

    FX.Privileges.GrantView.d: privileges_grantview_handler,
    FX.Privileges.GrantAppend.d: privileges_grantappend_handler,
    FX.Privileges.GrantHost.d: privileges_granthost_handler,
    FX.Privileges.GrantDiscover.d: privileges_grantdiscover_handler,
    FX.Privileges.GrantModifyRights.d: privileges_grantmodifyrights_handler,
    FX.Privileges.RevokeView.d: privileges_revokeview_handler,
    FX.Privileges.RevokeAppend.d: privileges_revokeappend_handler,
    FX.Privileges.RevokeHost.d: privileges_revokehost_handler,
    FX.Privileges.RevokeDiscover.d: privileges_revokediscover_handler,
    FX.Privileges.RevokeModifyRights.d: privileges_revokemodifyrights_handler,

    FX.Graph.Tag.d: graph_tag_handler,
    FX.Graph.Sync.d: graph_sync_handler,
    FX.Graph.TakeTransactorRole.d: graph_take_transactor_role_handler,
    FX.Graph.ReleaseTransactorRole.d: graph_release_transactor_role_handler,
    FX.Graph.Transact.d: graph_transaction_handler,
    FX.Graph.Load.d: graph_load_handler,

    
    FX.Clipboard.CopyTo.d: clipboard_copy_to_handler,
    FX.Clipboard.CopyFrom.d: clipboard_copy_from_handler,

    FX.LocalFile.Read.d:  read_localfile_handler,
    FX.LocalFile.ReadBinary.d:  readbinary_localfile_handler,
    FX.LocalFile.Load.d:  load_localfile_handler,
    FX.LocalFile.Write.d: write_localfile_handler,
    FX.LocalFile.Save.d:  save_localfile_handler,
    FX.LocalFile.SystemOpenWith.d: system_open_with_handler,
    FX.LocalFile.MonitorPath.d: monitor_path_handler,

    FX.ZefHub.Login.d: zefhub_login_handler,
    FX.ZefHub.Logout.d: zefhub_logout_handler,

    FX.RAE.Tag.d: rae_tag_handler,
}



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import *
from ...ops import *

import json
import graphql


def parse_partial_graphql(schema):
    """Goes through the schema line by line and creates a json file that represents
    the schema one-to-one. This should be an invertible process.

    """
    doc = graphql.parse(schema)

    output = {
        "types": {},
        "enums": {}
    }

    # Find all Zef directives first, which are single-line comments that begin with # Zef.<name>:
    lines = (schema
             | split["\n"]
             | filter[lambda x: x.startswith("# Zef.")]
             | map[lambda x: x[len("# Zef."):]]
             | collect)

    for line in lines:
        parts = line.split(':', 1)
        assert len(parts) == 2, "Zef directive line contains no ':'."
        name, details = parts
        name = name.strip()

        if name == "SchemaVersion":
            assert "schema_version" not in output, "Not allowed to have multiple Zef.SchemaVersion directives."
            output["schema_version"] = details.strip()
            continue

        # Anything else requires json parsing
        try:
            details = json.loads(details)
        except Exception as exc:
            raise Exception(f"JSON parsing of Zef directive '{name}' failed") from exc

        if name == "Authentication":
            assert "auth" not in output, "Not allowed to have multiple Zef.Authentication directives."
            for key in details:
                assert key in ["Algo", "JWKURL", "Audience", "Header", "Namespace",
                               "VerificationKey", "VerificationKeyEnv", "Public"], f"Unknown auth key '{key}'"
            output["auth"] = details
        elif name == "Route":
            assert set(keys(details)) == {"route", "hook"}
            if "route" in ["/", "/gql"]:
                raise Exception("Custom route should not alias '/' or '/gql'")
            routes = output.setdefault("routes", [])
            routes.append((details["route"], details["hook"]))
        else:
            raise Exception(f"Unsupported Zef.{name} directive")

    if output.get("schema_version", None) != "v1":
        raise Exception(
            "Only support schemas with an explicit version of 'v1'. Please include:\n# Zef.SchemaVersion: v1\ninto your schema file.")

    for definition in doc.definitions:
        if definition.kind == "object_type_definition":
            t_def = output["types"][definition.name.value] = {}

            for directive in definition.directives:
                if directive.name.value == "auth":
                    if "auth" not in output:
                        raise Exception(
                            "Disallowing @auth directives when no # Zef.Authentication... line is present to prevent accidentally security holes.")
                    for arg in directive.arguments:
                        if arg.name.value == "query":
                            t_def["_AllowQuery"] = arg.value.value
                        elif arg.name.value == "add":
                            t_def["_AllowAdd"] = arg.value.value
                        elif arg.name.value == "update":
                            t_def["_AllowUpdate"] = arg.value.value
                        elif arg.name.value == "updatePost":
                            t_def["_AllowUpdatePost"] = arg.value.value
                        elif arg.name.value == "delete":
                            t_def["_AllowDelete"] = arg.value.value
                        else:
                            raise Exception(f"Don't know how to handle auth kind of {arg.name.value}")
                elif directive.name.value == "upfetch":
                    assert len(directive.arguments) == 1, "upfetch directive needs exactly one argument, 'field'"
                    arg = directive.arguments[0]
                    if arg.name.value != "field":
                        raise Exception("upfetch directive needs exactly one argument, 'field'")
                    t_def["_Upfetch"] = arg.value.value
                elif directive.name.value == "ET":
                    assert len(directive.arguments) == 1, "ET directive needs exactly one argument, 'et'"
                    arg = directive.arguments[0]
                    if arg.name.value != "et":
                        raise Exception("ET directive needs exactly one argument, 'et'")
                    t_def["_ET"] = arg.value.value
                elif directive.name.value == "hook":
                    for arg in directive.arguments:
                        for hook_name in ["Create", "Remove", "Update"]:
                            if arg.name.value == f"on{hook_name}":
                                t_def[f"_On{hook_name}"] = arg.value.value
                                break
                        else:
                            raise Exception(f"Unknown hook name {arg.name.value}")
                else:
                    raise Exception(f"Don't know how to handle type directive @{directive.name.value}")

            for field in definition.fields:
                f_def = t_def[field.name.value] = {}

                f_type = field.type
                if f_type.kind == "non_null_type":
                    f_def["required"] = True
                    f_type = f_type.type
                if f_type.kind == "list_type":
                    f_def["list"] = True
                    f_type = f_type.type
                if f_type.kind == "non_null_type":
                    f_def["listNonNullItem"] = True
                    f_type = f_type.type
                assert f_type.kind == "named_type", f"Don't know what this kind of field type is {f_type.kind}"
                f_def["type"] = f_type.name.value

                assert len(field.arguments) == 0, "Currently not allowing custom arguments to a field"
                for directive in field.directives:
                    v = directive.name.value
                    for bool_key in ["search", "unique", "incoming"]:
                        if v == bool_key:
                            f_def[bool_key] = True
                            assert len(directive.arguments) == 0, f"@{bool_key} cannot take arguments"
                            break
                    else:
                        if v == "relation":
                            args = parse_arguments_as_dict(directive.arguments)
                            assert args.keys() == {"rt"} or args.keys() == {"source", "rt", "target"}, "@relation must take either 'rt' or 'source', 'rt', 'target' arguments. Got {set(args.keys())}"
                            if "source" in args:
                                f_def["relation"] = (ET(args["source"]), RT(args["rt"]), ET(args["target"]))
                            else:
                                f_def["relation"] = RT(args["rt"])
                        elif v == "dynamic":
                            args = parse_arguments_as_dict(directive.arguments)
                            assert args.keys() == {"hook"}, "@dynamic must take a single argument, 'hook'"
                            f_def["dynamic"] = args["hook"]
                        else:
                            raise Exception(f"Don't know how to handle directive @{v}")
        elif definition.kind == "enum_type_definition":
            e_def = output["enums"][definition.name.value] = []
            assert len(definition.directives) == 0, "Don't accept directives on enums"

            for value in definition.values:
                e_def += [value.name.value]

                assert len(value.directives) == 0, "Don't accept directives on enum values"
        else:
            raise Exception(f"Don't know how to handle a top-level graphql schema object of kind {definition.kind}")

    return output


def string_to_RAET(s):
    if s.startswith("ET."):
        return ET(s[len("ET."):])
    elif s.startswith("RT."):
        return RT(s[len("RT."):])
    elif s.startswith("AET."):
        return AET(s[len("AET."):])
    elif s.startswith("EN."):
        return AET(s[len("EN."):])
    else:
        raise Exception(f"Can't parse string as a RAE type: {s}")


def parse_arguments_as_dict(args):
    d = {}
    for arg in args:
        d[arg.name.value] = arg.value.value
    return d


def simple_capitalize(s):
    # This is here to make things like firebaseID go to FirebaseID rather than
    # FirebaseId. This is more predicatable for users.
    assert len(s) > 0
    return s[0].upper() + s[1:]


def json_to_minimal_nodes(json, g):
    """This takes a json schema, such as one produced by the `parse_schema`
        function, and puts the minimal GQL schema nodes onto the graph.
        "Minimal" is the key word here, as the generated schema graph nodes will
        not define any extra resolvers, these will be done at server run-time.

    """

    core_types = {
        "String": "String",
        "Float": "Float",
        "DateTime": "Time",
        "Int": "Int",
        "Boolean": "Bool"
    }

    actions = []

    # The root node meta data
    actions += [ET.GQL_Root["root"]]
    if "auth" in json:
        # A bit of sanity checking here - we should have one algo, which selects
        # symmetric/asymmetric keys, which allows only particular details to be
        # given.
        for key in ["Algo", "Header", "Audience"]:
            assert key in json["auth"], f"Auth requires key '{key}'"
        algo = json["auth"]["Algo"]
        assert algo in ["RS256", "HS256"], f"Auth algo is '{algo}', must be one of 'RS256' or 'HS256'"

        actions += [(Z["root"], RT.AuthAlgo, algo)]
        actions += [(Z["root"], RT.AuthHeader, json["auth"]["Header"])]
        actions += [(Z["root"], RT.AuthAudience, json["auth"]["Audience"])]
        actions += [(Z["root"], RT.AuthPublic, json["auth"].get("Public", True))]
        if "Namespace" in json["auth"]:
            actions += [(Z["root"], RT.AuthNamespace, json["auth"]["Namespace"])]

        assert sum(x in json["auth"] for x in ["JWKURL", "VerificationKey", "VerificationKeyEnv"]) == 1, "Need exactly one of ['JWKURL', 'VerificationKey', 'VerificationKeyEnv'] in auth"
        if algo == "RS256":
            assert "JWKURL" in json["auth"], "RS256 needs JWKURL specified."
            actions += [(Z["root"], RT.AuthJWKURL, json["auth"]["JWKURL"])]
        if algo == "HS256":
            vkey = json["auth"].get("VerificationKey", None)
            if vkey is None:
                env = json["auth"].get("VerificationKeyEnv", None)
                if env is None:
                    raise Exception("HS256 needs either VerificationKey or VerificationKeyEnv specified.")
                import os
                if env not in os.environ:
                    raise Exception(f"VerificationKeyEnv asked for environment variable '{env}' which was not present.")
                vkey = os.environ[env]
            vkey = vkey.replace("\\n", '\n')
            actions += [(Z["root"], RT.AuthPresharedKey, vkey)]
    if "routes" in json:
        for route,hook in json["routes"]:
            if hook not in now(g):
                raise Exception(f"Hook named {hook} not found on schema graph")
            actions += [(Z["root"], RT.Route,
                         {ET.Route: {RT.Route: route,
                                     RT.Hook: g | now | get[hook] | collect}})]
                

    for gql_name,typ in core_types.items():
        actions += [
            getattr(AET, typ)[gql_name],
            # Helpful hack to make the resolvers easier
            (Z[gql_name], RT.Name, gql_name),
            (Z["root"], RT.GQL_CoreScalarType, Z[gql_name])
        ]

    def name_to_raet(name):
        if name in core_types:
            return getattr(internals.VRT, core_types[name])
        if name in json["enums"]:
            return getattr(internals.VRT.Enum, name)
        return ET(name)

    for type_name,fields in json["types"].items():

        actions += [(ET.GQL_Type[type_name], RT.Name, type_name)]
        actions += [(Z["root"], RT.GQL_Type, Z[type_name])]

        et_name = fields.get("_ET", type_name)
        actions += [(Z[type_name], RT.GQL_Delegate, delegate_of(ET(et_name)))]

        for field_name,field in fields.items():
            assert field_name != "id", "id is an automatically generated field, do not explicitly include"

            if field_name.startswith("_"):
                # Special handling here
                if field_name == "_ET":
                    continue
                elif field_name in ["_AllowQuery", "_AllowAdd", "_AllowUpdate", "_AllowUpdatePost", "_AllowDelete"]:
                    # TODO: Turn into a zef function later on
                    actions += [(Z[type_name], RT(field_name[1:]), field)]
                elif field_name in ["_OnCreate", "_OnRemove", "_OnUpdaate"]:
                    # Find the zef function that this corresponds to
                    if field not in now(g):
                        raise Exception(f"Hook named {field} not found on schema graph")
                    z_func = g | now | get[field] | collect
                    actions += [(Z[type_name], RT(field_name[1:]), z_func)]

                elif field_name == "_Upfetch":
                    assert field in fields and "unique" in fields[field] and "required" in fields[field], f"Upfetch field '{field}' must exist, be unique and be required."
                    # We can figure out which relation we can attach the upfetch bool to.
                    qual_name = type_name + "__" + field
                    actions += [(Z[qual_name], RT.Upfetch, True)]
                else:
                    raise Exception(f"Don't understand special field name '{field_name}'")
            else:
                field = {**field}

                qual_name = type_name + "__" + field_name
                actions += [
                    (Z[type_name], RT.GQL_Field[qual_name], Z[field["type"]]),
                    (Z[qual_name], RT.Name, field_name)
                ]
                for bool_key in ["search", "unique", "incoming", "list", "required", "listNonNullItem"]:
                    if bool_key in field:
                        actions += [(Z[qual_name], RT(simple_capitalize(bool_key)), field[bool_key])]
                        del field[bool_key]

                if "dynamic" in field:
                    assert "relation" not in field, f"Can't give a relation with a dynamically resolved field: {field}"
                    if field["dynamic"] not in now(g):
                        raise Exception(f"Hook named {field['dynamic']} not found on schema graph")
                    z_func = g | now | get[field["dynamic"]] | collect
                    actions += [(Z[qual_name], RT.GQL_FunctionResolver, z_func)]
                    del field["dynamic"]
                else:
                    if "relation" in field:
                        resolve_with = delegate_of(field["relation"])
                        del field["relation"]
                    else:
                        this = ET(type_name)
                        rt = RT(simple_capitalize(field_name))
                        other = name_to_raet(field["type"])
                        if field.get("incoming", False):
                            resolve_with = delegate_of((other, rt, this))
                        else:
                            resolve_with = delegate_of((this, rt, other))

                    actions += [(Z[qual_name], RT.GQL_Resolve_With, resolve_with)]

                del field["type"]
                assert len(field) == 0, f"There are unhandled keys left in field object {field}"

    for enum_name,opts in json["enums"].items():
        actions += [(ET.GQL_Enum[enum_name], RT.Name, enum_name)]
        actions += [(Z["root"], RT.GQL_Enum, Z[enum_name])]

        for opt in opts:
            actions += [(Z[enum_name], RT.GQL_Field, EN(enum_name, opt))]

    return actions






# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported

import zef
from zef import *
from zef.ops import *


class MyTestCase(unittest.TestCase):
    def test_rel_delegate_from_rel_source_or_target(self):
        g = Graph()

        I = lambda *args: instantiate(*args, g)
        with Transaction(g):
            m1 = I(ET.Machine)
            m2 = I(ET.Machine)
            r1 = I(m1, RT.EtoE, m2)
            rel_E_DR = I(m1, RT.EtoDR, r1 | delegate_of | collect)
        zef.pyzef.verification.verify_graph_double_linking(g.graph_data)
        # serialize_check(g)




if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...core.logger import log
log.warning("Use of deprecated module zef.gql.simplegql. Please use zef.graphql.simplegql instead")
log.warning("Use of deprecated module zef.gql.simplegql. Please use zef.graphql.simplegql instead")
log.warning("Use of deprecated module zef.gql.simplegql. Please use zef.graphql.simplegql instead")

from ...graphql.simplegql import __main__


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import partial
from .value_type import *


# the following functions add a layer of indirection to prevent circular imports

def graph_ctor(*args, **kwargs):
    # Go to experimental load path for cases it supports
    if len(kwargs) == 0 and len(args) == 1 and type(args[0]) == str:
        from ...experimental.repl_interface import load_graph
        return load_graph(args[0])
    else:
        from ... import core
        return core.Graph(*args, **kwargs)



def time_ctor(*args, **kwargs):
    from ... import pyzef
    return pyzef.main.Time(*args, **kwargs)


def bytes_ctor(*args, **kwargs):
    # from ... import core
    from ...core.bytes import Bytes_
    # return core.bytes.Bytes_(*args, **kwargs)
    return Bytes_(*args, **kwargs)


def decimal_ctor(*args, **kwargs):
    # from ... import core
    from ...core.decimal import Decimal_
    # return core.bytes.Bytes_(*args, **kwargs)
    return Decimal_(*args, **kwargs)


def flatgraph_ctor(*args, **kwargs):
    from ...core.flat_graph import FlatGraph_
    return FlatGraph_(*args)


def union_getitem(x):
    from ..op_structs import ZefOp
    if isinstance(x, tuple):
        return ValueType_(type_name='Union', absorbed=x)
    elif isinstance(x, ValueType_):
        return ValueType_(type_name='Union', absorbed=(x,))
    elif isinstance(x, ZefOp):
        return ValueType_(type_name='Union', absorbed=(x,))
    else:
        raise Exception(f'"Union[...]" called with unsupported type {type(x)}')

def intersection_getitem(x):
    from ..op_structs import ZefOp
    if isinstance(x, tuple):
        return ValueType_(type_name='Intersection', absorbed=x)
    elif isinstance(x, ValueType_):
        return ValueType_(type_name='Intersection', absorbed=(x,))
    elif isinstance(x, ZefOp):
        return ValueType_(type_name='Intersection', absorbed=(x,))
    else:
        raise Exception(f'"Intersection[...]" called with unsupported type {type(x)}')

def complement_getitem(x):
    if isinstance(x, ValueType_):
        return ValueType_(type_name='Complement', absorbed=(x,))
    else:
        raise Exception(f'"Complement[...]" called with unsupported type {type(x)}')

def is_getitem(x):
    from typing import Callable
    from ..op_structs import ZefOp
    from .. import func
    if isinstance(x, tuple):
        return ValueType_(type_name='Is', absorbed=x)
    elif isinstance(x, ValueType_):
        return ValueType_(type_name='Is', absorbed=(x,))
    elif isinstance(x, ZefOp):
        return ValueType_(type_name='Is', absorbed=(x,))
    elif isinstance(x, Callable):
        return ValueType_(type_name='Is', absorbed=(func[x],))
    else:
        raise Exception(f'"Is[...]" called with unsupported type {type(x)}')



def setof_ctor(*args):
    """
    Can be called with either SetOf[5,6,7] or SetOf(5,6,7).
    When calling with square brackets, a tuple must always be 
    passed implicitly or explicitly.
    SetOf[42] is NOT valid
    SetOf[(42,)] is valid
    SetOf[42,] is valid    
    SetOf[42, 43] is valid    
    """
    return ValueType_(type_name='SetOf', absorbed = (args, ))



def setof_getitem(x):
    if not isinstance(x, tuple):
        raise TypeError(f"`SetOf[...]` must be called with a tuple, either explicitly or implicitly. e.g. SetOf[3,4], SetOf[(3,4)]. When wrapping one value, use SetOf[42,]. It was called with {x}. ")
    return ValueType_(type_name='SetOf', absorbed=(x, ))




def rp_getitem(x):
    if not isinstance(x, tuple) or len(x)!=3:
        raise TypeError(f"`RP`[...]  must be initialized with a triple to match on. Got x={x}")
    return ValueType_(type_name='RP', absorbed=(x,))


def same_as_get_item(y):
    from ..op_implementations.implementation_typing_functions import discard_frame_imp
    return Is[lambda x: discard_frame_imp(x) == discard_frame_imp(y)]



Nil        = ValueType_(type_name='Nil',        constructor_func=None)
Any        = ValueType_(type_name='Any',        constructor_func=None)
Bool       = ValueType_(type_name='Bool',       constructor_func=bool)
Int        = ValueType_(type_name='Int',        constructor_func=int)
Float      = ValueType_(type_name='Float',      constructor_func=float)
String     = ValueType_(type_name='String',     constructor_func=str)
Bytes      = ValueType_(type_name='Bytes',      constructor_func=bytes_ctor)
Decimal    = ValueType_(type_name='Decimal',    constructor_func=decimal_ctor)
List       = ValueType_(type_name='List',       constructor_func=tuple)
Dict       = ValueType_(type_name='Dict',       constructor_func=dict)
Set        = ValueType_(type_name='Set',        constructor_func=set)
EZefRef    = ValueType_(type_name='EZefRef',    constructor_func=None)    
ZefRef     = ValueType_(type_name='ZefRef',     constructor_func=None)     
Graph      = ValueType_(type_name='Graph',      constructor_func=graph_ctor)     
GraphSlice = ValueType_(type_name='GraphSlice', constructor_func=None)     
FlatGraph  = ValueType_(type_name='FlatGraph',  constructor_func=flatgraph_ctor)     
ZefOp      = ValueType_(type_name='ZefOp',      constructor_func=None)     
Stream     = ValueType_(type_name='Stream',     constructor_func=None)     
TX         = ValueType_(type_name='TX',         constructor_func=None)     
Time       = ValueType_(type_name='Time',       constructor_func=time_ctor)
Error      = ValueType_(type_name='Error',      constructor_func=None)
Image      = ValueType_(type_name='Image',      constructor_func=None)
ValueType  = ValueType_(type_name='ValueType',  constructor_func=ValueType_)
T          = ValueType_(type_name='T',  constructor_func=None)
T1         = ValueType_(type_name='T1',  constructor_func=None)
T2         = ValueType_(type_name='T2',  constructor_func=None)
T3         = ValueType_(type_name='T3',  constructor_func=None)


# QuantityInt= ValueType_(type_name='QuantityInt',constructor_func=None)
# QuantityFloat= ValueType_(type_name='QuantityFloat',constructor_func=None)

UID        = ValueType_(type_name='UID',        constructor_func=None)       
BaseUID    = ValueType_(type_name='BaseUID',    constructor_func=None)       
ZefRefUID  = ValueType_(type_name='ZefRefUID',  constructor_func=None)        
EternalUID = ValueType_(type_name='EternalUID', constructor_func=None)       

Instantiated = ValueType_(type_name='Instantiated', constructor_func=None)       
Terminated = ValueType_(type_name='Terminated', constructor_func=None)       
Assigned   = ValueType_(type_name='Assigned',   constructor_func=None)       
Tagged     = ValueType_(type_name='Tagged',     constructor_func=None)       

LazyValue  = ValueType_(type_name='LazyValue',  constructor_func=None) 
Awaitable  = ValueType_(type_name='Awaitable',  constructor_func=None) 



AET        = ValueType_(type_name='AET',        constructor_func=None)
ET         = ValueType_(type_name='ET',         constructor_func=None)
RT         = ValueType_(type_name='RT',         constructor_func=None)
BT         = ValueType_(type_name='BT',         constructor_func=None)
Enum       = ValueType_(type_name='Enum',       constructor_func=None)
Tuple      = ValueType_(type_name='Tuple',     constructor_func=None)       
Function   = ValueType_(type_name='Function',   constructor_func=None)
GraphDelta = ValueType_(type_name='GraphDelta', constructor_func=None)     
Query      = ValueType_(type_name='Query',      constructor_func=None) 
Effect     = ValueType_(type_name='Effect',     constructor_func=None) 
DataFrame  = ValueType_(type_name='DataFrame',  constructor_func=None)     
# EZefRefs   = ValueType_(type_name='EZefRefs', constructor_func=None)     
# EZefRefss  = ValueType_(type_name='EZefRefss', constructor_func=None)     
# ZefRefs    = ValueType_(type_name='ZefRefs', constructor_func=None)     
# ZefRefss   = ValueType_(type_name='ZefRefss', constructor_func=None)     

Pattern        = ValueType_(type_name='Pattern',             constructor_func=None)
Union          = ValueType_(type_name='Union',               constructor_func=None,             get_item_func=union_getitem)
Intersection   = ValueType_(type_name='Intersection',        constructor_func=None,             get_item_func=intersection_getitem)
Is             = ValueType_(type_name='Is',                  constructor_func=None,             get_item_func=is_getitem)
SetOf          = ValueType_(type_name='SetOf',               constructor_func=setof_ctor,       get_item_func=setof_getitem)
Complement     = ValueType_(type_name='Complement',          constructor_func=None,             get_item_func=complement_getitem)
RP             = ValueType_(type_name='RP',                  constructor_func=None,             get_item_func=rp_getitem)
HasValue       = ValueType_(type_name='HasValue',            constructor_func=None)
SameAs         = ValueType_(type_name='SameAs',                                                 get_item_func = same_as_get_item)




def operates_on_ctor(x):
    from .._ops import operates_on, contains
    return Is[operates_on | contains[x]]

def related_ops_ctor(x):
    from .._ops import related_ops, contains
    return Is[related_ops | contains[x]]

def used_for_ctor(x):
    from .._ops import used_for, contains
    return Is[used_for | contains[x]]

OperatesOn     = ValueType_(type_name='OperatesOn',   constructor_func = operates_on_ctor,  get_item_func = operates_on_ctor)
RelatedOps     = ValueType_(type_name='RelatedOps',   constructor_func = related_ops_ctor,  get_item_func = related_ops_ctor)
UsedFor        = ValueType_(type_name='UsedFor',      constructor_func = used_for_ctor, get_item_func =used_for_ctor)



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *


class MyTestCase(unittest.TestCase):
    def test_do_not_show_terminated_relents_in_Transaction(self):
        g = Graph()
        with Transaction(g) as ctx: 
            m1 = instantiate(ET.Machine, g)
            m2 = instantiate(ET.Machine, g)
            m3 = instantiate(ET.Machine, g)
        terminate(m2) | g | run
        m4 = instantiate(ET.Machine, g)

        self.assertEqual(m4 | frame | collect, g | now | collect)
        ts4 = m4 | frame | graph_slice_index | collect

        with Transaction(g) as ctx: 
            this_frame = frame(ctx)
            self.assertEqual(this_frame, g | now | collect)
            self.assertEqual(ctx | to_graph_slice | graph_slice_index | collect, g | now | graph_slice_index | collect)
            self.assertEqual(ctx | to_graph_slice | graph_slice_index | collect, ts4+1)
            

        mm1 = m1 | to_ezefref | to_frame[g | now|collect] | collect
        mm1d = m1 | now | collect
        self.assertEqual(mm1, mm1d)
        
        self.assertEqual(m1 | now | frame | collect, g|now | collect)

        # self.assertEqual(ZefRefs([m1, m3]) | to_frame[now(g)] | frame | collect, g|now | collect)
        # TODO: Get to_frame working on ZefRefs again
        # self.assertEqual(ZefRefs([m1, m3]) | to_frame[g | now|collect] | frame | collect, g|now | collect)

        




if __name__ == '__main__':
    unittest.main()



# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class Decimal_:
    def __init__(self, val, decimal_positions):
        if not isinstance(decimal_positions, int):
            raise ValueError(f"in constructing a decimal: decimal_positions must be an int. got {type(decimal_positions)}")
        
        if isinstance(val, float):
            self.val_int_encoded = round(val * 10**decimal_positions)
            self.decimal_positions = decimal_positions
        elif isinstance(val, str):
            pos = val.find('.')
            pre = val if pos==-1 else val[:pos]
            post = (f"{{:0<{decimal_positions}}}").format('' if pos==-1 else val[pos+1:pos+decimal_positions+1])
            # check whether we need to round up
            m = decimal_positions+2
            postpost = val[m:m+1]
            rounding_term = 1 if len(postpost)==1 and int(postpost)>4 else 0
            self.val_int_encoded = int(pre+post) + rounding_term
            self.decimal_positions = decimal_positions
        else:
            raise ValueError(f"Can't construct a Decimal from a {type(val)}")

    def __repr__(self):
        # correct for repr output if decimal starts with '0': pad str repr of int with zeros explicitly        
        s=(f"{{:0>{self.decimal_positions}}}").format(str(self.val_int_encoded))
        return f"Decimal({s[:-self.decimal_positions]}.{s[-self.decimal_positions:]}, {self.decimal_positions})"

    def __str__(self):
        s=(f"{{:0>{self.decimal_positions}}}").format(str(self.val_int_encoded))
        return f"{s[:-self.decimal_positions]}.{s[-self.decimal_positions:]}"

    def __float__(self):
        return float(self.val_int_encoded / 10**self.decimal_positions)
    
    def __add__(self, other):
        if isinstance(other, int):
            return Decimal_((self.val_int_encoded+ other * 10**self.decimal_positions)/10**self.decimal_positions , self.decimal_positions)
        elif isinstance(other, float):
            return Decimal_((self.val_int_encoded + round(other * 10**self.decimal_positions)) /10**self.decimal_positions, self.decimal_positions)
        elif isinstance(other, Decimal_):
            """
            Use the higher decimal precision of the two operands for the output
            """
            # pos_max = max(self.decimal_positions, other.decimal_positions)
            # pos_min = min(self.decimal_positions, other.decimal_positions)
            
            # return Decimal_((self.val_int_encoded + round(other.val_int_encoded * 10**self.decimal_positions)) 
            #  /10**pos_max, pos_max)
            raise NotImplementedError
        else:
            raise TypeError
    
    def __radd__(self, other):
        return self+other
    
    def __sub__(self, other):
        return self+ (-1*other)
    
    def __rsub__(self, other):
        # :(  very mutaty to avoid unnecessary multiplications and divisions
        tmp = Decimal_(0.0, self.decimal_positions)
        tmp.val_int_encoded = -self.val_int_encoded
        return tmp + other

    def __neg__(self):
        """
        allow writing "-my_decimal"
        """
        tmp = Decimal_(0.0, self.decimal_positions)
        tmp.val_int_encoded = -self.val_int_encoded
        return tmp

    def __pos__(self):
        """
        allow writing "+my_decimal"
        """
        return self
    
    def __mul__(self, x):
        if not isinstance(x, int):
            raise NotImplementedError('You can only multiply Decimal with integers for now')
        tmp = Decimal_(0.0, self.decimal_positions)
        tmp.val_int_encoded = x*self.val_int_encoded
        return tmp

    def __rmul__(self, x):
        return self*x




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._ops import *
from .._core import *

##############################
# * Unpack
#----------------------------

def unpack_implementation(inputs, f):
    """
    used to wrap an zefop or a function. If the function
    takes multiple arguments, it is converted into a function
    that takes one tuple or list as a single argument.
    This list is unpacked into the function arguments.

    ---- Signature ----
    (Tuple[T1, T2, ...], Func[T1, T2, ...][T] ) -> T

    ---- Examples ----
    >>> (42, 10) | unpack[subtract]      # => 32


    ---- Tags ----
    used for: control flow
    related zefop: reverse_args
    """
    from typing import Generator
    if isinstance(inputs, tuple) or isinstance(inputs, list):
        return f(*inputs)
    if isinstance(inputs, Generator):
        return f(*tuple(inputs))

    elif isinstance(inputs, dict):
        # if this has wrapped a call[...] then we need to unpack the function
        # inside of it, as call itself does not handle kwargs

        # Why is this nesting so confusing???
        while isinstance(f, ZefOp):
            assert len(f.el_ops) == 1 and f.el_ops[0][0] == RT.Call, "Can only support a single call[] zefop when using dictionary unpack"
            assert len(f.el_ops[0][1]) == 1, "Can't support curried arguments yet, e.g. unpack[func[something][1][2]], when using dicitonary unpack"
            f = f.el_ops[0][1][0]
        return f(**inputs)
    else:
        raise Exception("unpack requires either a tuple or a dictionary as input")

def unpack_type_info(op, curr_type):
    # TODO
    return VT.Any





# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ... import *
from ...ops import *
from .main import create_schema_graph
from .server2 import start_server

from ...core.logger import log

import os
import argparse
parser = argparse.ArgumentParser()
parser.add_argument("--schema-file", type=str, metavar="schema_file", default=os.environ.get("SIMPLEGQL_SCHEMA_FILE", None),
                    help="The graphql schema file.")
parser.add_argument("--data-tag", type=str, metavar="data_tag", default=os.environ.get("SIMPLEGQL_DATA_TAG", None),
                    help="The tag of the Zef graph with the data (not the schema).")
parser.add_argument("--scratch", action="store_true", default=(False if "SIMPLEGQL_SCRATCH" in os.environ else None),
                    help="Use a temporary disposable local graph instead of persisting it on ZefHub.")
parser.add_argument("--port", type=int, default=int(os.environ.get("SIMPLEGQL_PORT", "443")))
parser.add_argument("--bind", type=str, default=os.environ.get("SIMPLEGQL_BIND", "0.0.0.0"),
                    help="IP address to bind to.")
parser.add_argument("--no-host-role", action="store_false", dest="host_role", default=(False if "SIMPLEGQL_NO_HOST_ROLE" in os.environ else True),
                    help="Whether the server should acquire host role on the data graph.")
parser.add_argument("--create", action="store_true", default=(True if "SIMPLEGQL_CREATE" in os.environ else None),
                    help="If the data graph does not exist, then create a new blank graph.")
parser.add_argument("--hooks", type=str, dest="hooks_file", default=os.environ.get("SIMPLEGQL_HOOKS_FILE", None),
                    help="If present, a python file containing the hooks to make available for the schema file")
parser.add_argument("--init-hook", type=str, dest="init_hook", default=os.environ.get("SIMPLEGQL_INIT_HOOK", None),
                    help="If present, a function in the hooks file to run on initialisation of the data graph")
parser.add_argument("--debug-level", type=int, dest="debug_level", default=os.environ.get("SIMPLEGQL_DEBUG_LEVEL", "0"),
                    help="The amount of debug messages to output")
args = parser.parse_args()

if args.data_tag is not None and args.scratch:
    print("Can't specify both a data tag and scratch")
    raise SystemExit(4)
if args.data_tag is None and not args.scratch:
    print("Need one of data-tag or scratch specified.")
    raise SystemExit(4)


schema_gql = args.schema_file | read_file | run | get["content"] | collect
if args.hooks_file is not None:
    hooks_string = args.hooks_file | read_file | run | get["content"] | collect
    log.info(f"Going to read hooks from '{args.hooks_file}'")
else:
    hooks_string = None

root = create_schema_graph(schema_gql, hooks_string)
log.info(f"Created schema graph from '{args.schema_file}'")

if args.data_tag is not None:
    guid = lookup_uid(args.data_tag)
    if guid is None:
        if not args.create:
            print("Graph is not known to us, and create is False. Exiting.")
            raise SystemExit(2)
        g_data = Graph(True)
        try:
            g_data | tag[args.data_tag] | run
        except:
            print("Unable to create and tag graph with '{args.data_tag}'. Maybe this tag is already taken by another user?")
            raise SystemExit(2)
        log.info("Created data graph with tag", tag=args.data_tag)
    else:
        g_data = Graph(args.data_tag)
        log.info("Loaded existing data graph")
else:
    assert args.scratch
    g_data = Graph()
    log.info("Created blank scratch data graph")

if args.host_role and not args.scratch:
    try:
        g_data | take_transactor_role | run
    except:
        print("""
Unable to obtain host role for data graph. Either stop other processes from having host role on this graph (`g | release_transactor_role | run` in that process) or run this server with `--no-host-role`).
        
Note that running without the host role is currently dangerous as mutations do not currently verify pre-conditions.
""") 
        raise SystemExit(3)
    log.info("Obtained host role on data graph")

if args.init_hook is not None:
    g_schema = Graph(root)
    if args.init_hook not in now(g_schema):
        log.error("Could not find init hook on schema graph", init_hook=args.init_hook)
        raise SystemExit(4)
    hook_func = g_schema | now | get[args.init_hook] | collect
    hook_func(g_data)
    log.info(f"Ran init hook: {args.init_hook}")
    

server_uuid = start_server(root, g_data,
                           port=args.port,
                           bind_address=args.bind,
                           debug_level=args.debug_level)

import time
try:
    while True:
        time.sleep(1)
except KeyboardInterrupt:
    pass
finally:
    {
        "type": FX.HTTP.StopServer,
        "server_uuid": server_uuid
    } | run


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from . import *
from ..core import *
from ..ops import *

# -----------------------Table View------------------------------------------
def generate_mapping_for_et(et, g):
    z = delegate_of(et, g)
    connected_rels = z | out_rels[RT] | collect
    mapping =  connected_rels | map[rae_type] | collect
    return mapping

@func
def value_or_type(node):
    if is_a(node, AET):
        val =  node | value | collect
        if isinstance(val, bool):
            return ["❌","✅"][val]
        return val
    else:
        default = f"({repr(rae_type(node))})"
        if len(Outs(node, RT.Name)) == 1:
            return f"{node | Out[RT.Name]  | value | collect}{default}"
        
        return default

@func
def traverse_rt(rt, et, traverse_op = Outs):
    outs_or_ins = et | traverse_op[rt] | collect
    if len(outs_or_ins) == 1:
        return outs_or_ins[0] | value_or_type | func[str] | collect
    elif len(outs_or_ins) > 1:
        return outs_or_ins | map[value_or_type | func[str]] | join[','] | collect
    else:
        return "-" 

def generate_rows(nodes, mapping, et_type = None):
    if et_type: return nodes | map[lambda et: mapping | map[traverse_rt[et]] | prepend[et_type] | func[tuple] | collect] | collect
    return nodes | map[lambda et: mapping | map[traverse_rt[et]] | func[tuple] | collect] | collect


@func
def generate_table_for_single_type_zrs(selected_et, zrs, expand, limit):
    g = Graph(zrs[0])
    mapping = generate_mapping_for_et(selected_et, g)
    mapping_values = [str(m) for m in mapping]

    zrs = zrs | take[limit] | collect
    padding = [(0,0,0,0),(1,1,1,1)][expand]
    

    colors = ["#ff7e74","#ffe596","#81d76d","#57acf9","#baaee1"]
    row_styles  = ["", "dim"]

    rows = generate_rows(zrs, mapping)
    columns = [ Column(Text(c,justify = "center"), header_style = Style(background_color="#2a3240", color = colors[i%len(colors)]), style = Style(color = colors[i%len(colors)]))
                for i,c in enumerate(mapping_values)]

    title = Frame(Text(f" A List of {len(zrs)} {repr(selected_et)}", bold= True, justify="center"), box ="simple_head")
    table = Table(
        show_edge=False,
        expand=True,     
        padding = padding,
        show_header=True,
        box = 'simple_head',        
        rows = rows,
        cols = columns,
        row_styles=row_styles,
    )  

    return VStack([
        title,
        table
    ])

@func
def generate_table_for_zrs(groups, expand, limit):
    if len(groups) == 0: return Frame()
    if len(groups) == 1: return generate_table_for_single_type_zrs(*groups[0], expand, limit)


    g = Graph(groups[0][1][0])
    mapping = groups | map[lambda tup: generate_mapping_for_et(tup[0], g)] | concat  | collect
    
    padding = [(0,0,0,0),(1,1,1,1)][expand]

    colors = ["#ff7e74","#ffe596","#81d76d","#57acf9","#baaee1"]
    row_styles  = ["", "dim"]

    rows =  groups | map[lambda tup: generate_rows(tup[1], mapping, str(tup[0]))] | concat | take[limit] | collect

    mapping = ["ET"] + mapping
    mapping_values = [str(m) for m in mapping]
    columns = [ Column(Text(c,justify = "center"), header_style = Style(background_color="#2a3240", color = colors[i%len(colors)]), style = Style(color = colors[i%len(colors)])) for i,c in enumerate(mapping_values)]

    title = Frame(Text(f"Table of {','.join([str(g[0]) for g in groups])}", bold= True, justify="center"), box ="simple_head")
    table = Table(
        show_edge=False,
        expand=True,     
        padding = padding,
        show_header=True,
        box = 'simple_head',        
        rows = rows,
        cols = columns,
        row_styles=row_styles,
    )  

    return VStack([
        title,
        table
    ])

@func
def generate_table_from_query(query, expand = False, limit=20):
    groups = query | filter[is_a[ET]] | group_by[rae_type] |  collect
    table = generate_table_for_zrs(groups, expand, limit)
    return VStack([table])


#------------------------Card View------------------------------
def generate_rows_for_node(node, is_out = True):
    traverse_op, traverse_op_end = [(in_rels, Ins), (out_rels, Outs)][is_out]
    return node | traverse_op[RT] | map[rae_type] | func[set] | map[lambda rt: (str(rt), traverse_rt(rt, node, traverse_op_end))] | collect

def generate_rts_table(zr, is_out, expand):
    colors = ["#ff7e74","#ffe596","#81d76d","#57acf9","#baaee1"]
    padding = [(0,0,0,0),(1,1,1,1)][expand]


    rows = generate_rows_for_node(zr, is_out)
    if not rows: return ""
    row_styles  = [colors[i%len(colors)] for i in range(len(rows))]

    column_headers = ["", [" Ins 👈🏻", " Outs 👉🏻"][is_out]]
    columns = [Column(Text(c,justify = "left", style = Style(color = colors[1]))) for i,c in enumerate(column_headers)]

    return Table(
        expand=True,     
        padding= padding,
        box = 'simple_head',        
        rows = rows,
        cols = columns,
        row_styles=row_styles,
    )  

@func
def generate_card(zr, expand = False):
    title = Text(f"{(repr(rae_type(zr)))}", bold= True)
    subtitle = Text(f"({uid(zr)})", bold= True)

    outs_table = generate_rts_table(zr, True, expand)
    ins_table = generate_rts_table(zr, False, expand)

    return Frame(HStack([outs_table, ins_table], expand= True), title = title, subtitle=subtitle,  box="horizontals") 


#--------------------------Ops--------------------------------------
to_table =  generate_table_from_query
to_card  =  generate_card


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ..core import *
from ..ops import *

import pandas
import math
from caseconverter import pascalcase

def is_nan(val):
    import pandas.core.dtypes.common as p_dtypes
    if p_dtypes.is_integer(val):
        return math.isnan(val)
    elif p_dtypes.is_float(val):
        return math.isnan(val)
    else:
        return False

def get_col_ind(cols, name):
    for i,col in enumerate(cols):
        if col["name"] == name:
            return i
    return None

def get_col(cols, name):
    return cols[get_col_ind(cols,name)]
        

def process_df_guess(df, tag):
    out = {
        "tag": tag,
        "cols": [],
    }

    cols = list(df)
    df2 = df.convert_dtypes()

    for col in cols:

        sw = df2[col].dtype
        import pandas.core.dtypes.common as p_dtypes
        if p_dtypes.is_string_dtype(sw):
            typ = 'String'
        elif p_dtypes.is_bool_dtype(sw):
            typ = 'Bool'
        elif p_dtypes.is_integer_dtype(sw):
            typ = 'Int'
        elif p_dtypes.is_float_dtype(sw):
            typ = 'Float'
        else:
            typ = 'Serialized'
        
        # TODO: We could put a guess here on cols which are ints and have "ID/Id" at the end of them.
        col_desc = {
            "name": col,
            "data_type": typ,
            # Everything starts off as a field
            "purpose": "field",
            "RT": pascalcase(col),
        }

        out["cols"] += [col_desc]

    col_names = out["cols"] | map[get["name"]] | collect
    if len(distinct(col_names)) != len(col_names):
        raise Exception("CSV has multiple columns with the same name")

    ID_col = None
    # The first column that is an integer is a good candidate for the ID
    for col in out["cols"]:
        if col["data_type"] == 'Int':
            ID_col = col["name"]
            break
    if ID_col is None:
        # Otherwise try the first string column
        for col in out["cols"]:
            if col["data_type"] == 'String':
                ID_col = col["name"]
                break

    if ID_col is None:
        raise Exception("Need an ID column!")

    if False:
        pass
    else:
        # Assume this table defines the fields of an entity
        out["kind"] = "entity"
        out["ET"] = pascalcase(tag)
        out["ID_col"] = ID_col

        col = get_col(out["cols"], ID_col)
        new_col = {
            "name": col["name"],
            "data_type": col["data_type"],
            "purpose": "id",
            "RT": "ID"
        }
        col.clear()
        col.update(new_col)

    return out


def guess_csvs(filenames):
    if isinstance(filenames, str):
        import glob
        filenames = glob.glob(filenames)

    defs = []
    for filename in filenames:
        tag = os.path.splitext(filename)[0]
        try:
            df = pandas.read_csv(filename)
        except pandas.errors.EmptyDataError:
            print(f"Pandas complained that file {filename!r} has no data - going to skip")
            continue
        except Exception as exc:
            raise Exception(f"Pandas couldn't parse file {filename!r}") from exc
        this = process_df_guess(df, tag)

        this["data_source"] = {
            "type": "csv",
            "filename": filename
        }

        defs += [this]
    
    decl = default_outer()
    decl["definitions"] = defs
    return decl

def default_outer():
    return {
        "default_ID": "ID"
    }

def et_id(et, val, decl):
    ID_col = get_ent_ID_col(et, decl)
    val = coerce_val(val, aet_str_to_aet(ID_col['data_type']))
    assert val is not None and not is_nan(val)
    return f"<ET_{et} {val}>"

def rt_id(src, trg, rt, val, data_type):
    val = coerce_val(val, aet_str_to_aet(data_type))
    assert val is not None and not is_nan(val)
    return f"<RT_{src}_{trg}_{rt} {val}>"

def aet_str_to_aet(aet_s):
    parts = aet_s.split('.')
    aet = AET
    for part in parts:
        aet = getattr(aet, part)
    return aet

def coerce_val(val, aet):
    if isinstance(val, list):
        raise Exception("Vals can't be lists")

    if is_a(aet, AET.QuantityFloat):
        return QuantityFloat(val, aet.__unit)
    elif is_a(aet, AET.QuantityInt):
        return QuantityFloat(val, aet.__unit)
    elif is_a(aet, AET.Enum):
        return EN(aet.__enum_type, to_pascal_case(val))
    elif is_a(aet, AET.Bool):
        if val in [0, False, 'False', 'false', 'FALSE', 'no']:
            return False
        elif val in [1, True, 'True', 'true', 'TRUE', 'yes']:
            return True
        else:
            raise Exception(f"Bool value can't be converted: {val}")
    elif is_a(aet, AET.Time):
        if isinstance(val, str):
            # If all numeric chars are zeros, this is a null time
            null = distinct([x for x in val if x.isdigit()]) == ['0']
            if null:
                return None
            return Time(val)
        else:
            raise Exception("Don't know how to convert to AET.Time")
    # elif is_a(aet, str):
    #     pass
    # else:
    #     assert not math.isnan(val)
    return val

def get_ent_ID_col(et, decl):
    # Try to find et in the definitions - if not there, then we'll just use ID
    # as the ID field
    for d in decl["definitions"]:
        if d.get("ET", None) == et:
            return get_col(d["cols"], d["ID_col"])
    fake_col = {
        "RT": decl["default_ID"],
        "data_type": "Int",
    }
    return fake_col

def make_ent(col, val, decl):
    et = col["ET"]

    z_id = et_id(et, val, decl)
    actions = [ET(et)[z_id]]

    ID_col = get_ent_ID_col(et, decl)
    actions += attach_field(Z[z_id], ID_col, val)

    return actions

def attach_field(z, col, val, rt=None):
    import math
    if val is None or isinstance(val, float) and math.isnan(val):
        return []
    
    if rt is None:
        rt = RT(col["RT"])
    aet = aet_str_to_aet(col['data_type'])
        
    ae_id = name_from_Z(z) + f" field {rt}"
    val = coerce_val(val, aet)
    if val is None:
        return []

    return [
        aet[ae_id],
        Z[ae_id] <= val,
        (z, rt, Z[ae_id]),
    ]

def name_from_Z(z):
    return LazyValue(z) | peel | first | second | first | collect

def import_actions_definition(definition, decl):
    assert validate_definition(definition, decl)
    actions = []
    
    if definition["data_source"]["type"] == "csv":
        df = pandas.read_csv(definition["data_source"]["filename"])
    else:
        raise Exception("Unknown data source type")

    groups = definition["cols"] | group_by[get["purpose"]][["entity","field", "field_on", "ignore", "id", "source", "target"]] | func[dict] | collect

    # col_ents = definition["cols"] | filter[
    #     get["purpose"] | equals["entity"]
    # ] | collect

    # col_fields = definition["cols"] | filter[
    #     get["purpose"] | equals["field"]
    # ] | collect
    col_ents = groups["entity"]
    col_fields = groups["field"]
    col_fieldons = groups["field_on"]

    if len(groups["id"]) == 0:
        col_ID = None
    elif len(groups["id"]) == 1:
        col_ID = only(groups["id"])

    # As this can be slow - let's warn about it
    next_print_time = now() + 5*seconds
    # Note: iterrows does implicit conversion of types. Use itertuples to maintain the dtypes
    # Also, itertuples can't allow python identifiers like "class".
    #
    # So they are both useless! Of course this is not recommended to do things
    # this way... but it doesn't help when I need to do things this way.
    #
    # Indices is the only way for now.
    for i_row in df.index:
        if now() > next_print_time:
            print(f"SQL import is taking some time. Currently up to row {i_row} of table tagged {definition['tag']}")
            next_print_time = now() + 5*seconds
            
        # First do everything that is not a relation
        for col_ent in col_ents:
            val = df[col_ent["name"]][i_row]
            val = coerce_val(val, aet_str_to_aet(col_ent["data_type"]))
            if val is not None and not is_nan(val):
                temp = make_ent(col_ent, val, decl)
                actions += temp

        # Next get the object which corresponds to this row
        if definition["kind"] == "entity":
            val = df[col_ID["name"]][i_row]
            this_id = et_id(definition["ET"], val, decl)
            actions += [ET(definition["ET"])[this_id]]
            z_this = Z[this_id]
        elif definition["kind"] == "relation":
            col_source = only(groups["source"])
            col_target = only(groups["target"])
            val_source = df[col_source["name"]][i_row]
            val_target = df[col_target["name"]][i_row]
            source_id = et_id(col_source["ET"], val_source, decl)
            target_id = et_id(col_target["ET"], val_target, decl)

            actions += [ET(col_source["ET"])[source_id]]
            actions += [ET(col_target["ET"])[target_id]]

            if col_ID is None:
                val = i_row
                this_id = rt_id(source_id, target_id, definition["RT"], val, "Int")
            else:
                val = df[col_ID["name"]][i_row]
                this_id = rt_id(source_id, target_id, definition["RT"], val, col_ID["data_type"])
            actions += [(Z[source_id], RT(definition["RT"])[this_id], Z[target_id])]
            z_this = Z[this_id]
        else:
            raise NotImplementedError()

        if col_ID is not None:
            val = df[col_ID["name"]][i_row]
            actions += attach_field(z_this, col_ID, val)

        # All fields for this object
        for col_field in col_fields:
            val = df[col_field["name"]][i_row]
            actions += attach_field(z_this, col_field, val)

        # All connections to the other entities
        for col_ent in col_ents:
            if col_ent["name"] == definition["ID_col"]:
                continue
            val = df[col_ent["name"]][i_row]
            val = coerce_val(val, aet_str_to_aet(col_ent["data_type"]))

            if val is not None and not is_nan(val):
                target_id = et_id(col_ent["ET"], val, decl)
                rel_id = f"{name_from_Z(z_this)} {col_ent['RT']} {target_id}"
                actions += [(z_this, RT(col_ent["RT"])[rel_id], Z[target_id])]

        # All fields on connections
        for col_fieldon in col_fieldons:
            col_target = get_col(col_ents, col_fieldon["target"])
            target_val = df[col_target["name"]][i_row]
            target_id = et_id(col_target["ET"], target_val, decl)
            rel_id = f"{name_from_Z(z_this)} {col_target['RT']} {target_id}"
            z_conn = Z[rel_id]
            val = df[col_fieldon["name"]][i_row]
            temp += attach_field(z_conn, col_fieldon, val)
            actions += temp

    actions = distinct(actions)
    return actions

def validate_definition(definition, decl):
    # if "ID_col" in definition:
    #     col_ID = get_col(definition["cols"], definition["ID_col"])
    #     if definition["kind"] == "entity":
    #         assert definition["ET"] == col_ID["ET"], f"The table definition of an entity needs to have the same ET set ('{definition['ET']}') for the column which is its identity ('{col_ID['ET']}')."

    return True

def import_actions(decl):
    actions = []
    for d in decl["definitions"]:
        actions += import_actions_definition(d, decl)
    return distinct(actions)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest  # pytest takes ages to run anything as soon as anything from zef is imported
from zef import *
from zef.ops import *

from zef.pyzef.zefops import on_value_assignment, subscribe, keep_alive, on_instantiation, on_termination, outgoing

class MyTestCase(unittest.TestCase):

    def test_subscribe(self):
        g = Graph()
        z = instantiate(AET.Int, g)

        on_value_assign_list = []
        def OnValueAssign(x, l):
            l += [x|value|collect]

        sub1 = z | subscribe[on_value_assignment][keep_alive[False]][lambda x: OnValueAssign(x, on_value_assign_list)]

        on_rel_list = []

        def OnInst(x, l):
            l += [("inst", RT(x))]
        def OnTerm(x, l):
            l += [("term", RT(x))]
        sub2 = z | subscribe[on_instantiation[outgoing][RT.Map]][keep_alive[False]][lambda x: OnInst(x, on_rel_list)]
        sub3 = z | subscribe[on_instantiation[outgoing][RT.Value]][lambda x: OnInst(x, on_rel_list)]
        sub4 = z | subscribe[on_termination[outgoing][RT.Map]][lambda x: OnTerm(x, on_rel_list)]
        sub5 = z | subscribe[on_termination[outgoing][RT.Value]][lambda x: OnTerm(x, on_rel_list)]

        z | assign[1] | g | run
        z | assign[2] | g | run
        z | assign[3] | g | run
        z | assign[4] | g | run
        z | assign[5] | g | run

        (z, RT.Map, 1) | g | run
        (z, RT.Map, 2) | g | run
        (z, RT.Value, 2) | g | run

        z | now | Outs[RT.Map] | map[terminate] | g | run

        (z, RT.Value, 3) | g | run
        (z, RT.Map, 4) | g | run

        z | now | Outs[RT.Value] | map[terminate] | g | run

        # Test that keep_alive kills subs when it is False only
        sub2 = None
        sub3 = None
        import gc
        gc.collect()

        (z, RT.Map, 10) | g | run
        (z, RT.Value, 10) | g | run

        # Test that unsubscribe kills subs definitely
        sub4.unsubscribe()
        z | now | Outs[RT.Map] | map[terminate] | g | run

        self.assertEqual(on_value_assign_list, [1,2,3,4,5])
        self.assertEqual(on_rel_list, [
            ("inst", RT.Map),
            ("inst", RT.Map),
            ("inst", RT.Value),
            ("term", RT.Map),
            ("term", RT.Map),
            ("inst", RT.Value),
            ("inst", RT.Map),
            ("term", RT.Value),
            ("term", RT.Value),
            ("inst", RT.Value),
        ])
                          
if __name__ == '__main__':
    unittest.main()




# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .resolvers_generator import generate_resolvers, fill_types_default_resolvers
from .schema_generator import generate_schema_str
from .schema_parser import generate_schema_dict

__all__ = [
    "generate_resolvers",
    "generate_schema_str",
    "generate_schema_dict",
    "make_graphql_api",
    "fill_types_default_resolvers"
]

def make_graphql_api(schema_dict: dict, g = None):
    """
    Given a schema dictionary, generate Ariadne's GraphQL Executable Schema.
    """
    from ariadne import make_executable_schema

    object_types = generate_resolvers(schema_dict, g)
    simple_schema = generate_schema_str(schema_dict)
    schema = make_executable_schema(simple_schema, object_types)
    return schema


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ...pyzef.zefops import (
    L,
    O,
    Q,
    affected,
    allow_terminated_relent_promotion,
    assert_length,
    batch,
    concatenate,
    delegate,
    element,
    exists_at,
    filter,
    first,
    flatten,
    has_in,
    has_out,
    has_relation,
    incoming,
    ins,
    ins_and_outs,
    instances,
    instances_eternal,
    instantiated,
    instantiation_tx,
    intersect,
    is_zefref_promotable,
    is_zefref_promotable_and_exists_at,
    keep_alive,
    last,
    lift,
    not_in,
    now,
    on_instantiation,
    on_termination,
    on_value_assignment,
    only,
    only_or,
    outgoing,
    outs,
    relation,
    set_union,
    sort,
    source,
    subscribe,
    take,
    target,
    terminate,
    terminated,
    termination_tx,
    time,
    time_slice,
    time_travel,
    to_frame,
    to_ezefref,
    tx,
    uid,
    unique,
    value,
    value_assigned,
    value_assignment_txs,
    without
)

from .attach import attach, fill_or_attach # this zefop is defined in python
make_field = attach # deprecation

# Need to override the filter behaviour because typing is hard in python
# from ..zef_functions import to_clipboard
# from ..graph_delta import transact

from ... import pyzef, core
pyzef.zefops.TimeZefopStruct.__getitem__ = lambda self, s: core.Time(s)


# This implementation is the imperative side and was moved from lazy_zefops
def _assign_value(z, value):
    from ..core import AET
    from ..ops import is_a
    from ..pyzef.zefops import assign_value, SerializedValue
    if is_a(z, AET.Serialized):
        from ..serialization import serialize, serializable_types
        if type(z) in serializable_types:
            from json import dumps
            value = SerializedValue("tools.serialize", dumps(serialize(value)))
        else:
            raise Exception(f"Don't know how to serialize a type of {val}")
    return assign_value(z, value)


# Copyright 2022 Synchronous Technologies Pte Ltd
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .._ops import *

def graph_take_transactor_role_handler(eff: dict):
    from ...pyzef.main import make_primary
    make_primary(eff["graph"], True)
    return {}

def graph_release_transactor_role_handler(eff: dict):
    from ...pyzef.main import make_primary
    make_primary(eff["graph"], False)
    return {}

def graph_transaction_handler(eff: dict):
    """[summary]

    Args:
        eff (dict): {
            "type": FX.Graph.Transact,
            "target_graph": g1,
            "commands": list_of_commands
        }
    """

    from ..graph_delta import perform_transaction_commands, filter_temporary_ids, unpack_receipt

    if eff["target_graph"].graph_data.is_primary_instance:
        receipt = perform_transaction_commands(eff['commands'], eff['target_graph'])
    else:
        from ..overrides import merge
        receipt = merge(eff["commands"], eff["target_graph"])
    
    # we need to forward this if it is in the effect
    if 'unpacking_template' in eff:
        return unpack_receipt(eff['unpacking_template'], receipt)

    receipt = filter_temporary_ids(receipt)
    return receipt
    

def graph_load_handler(eff: dict):
    # TODO: Make the underlying function be asynchronous and then update this to return an awaitable

    from ...pyzef.main import load_graph
    kwargs = {}
    if "mem_style" in eff:
        kwargs["mem_style"] = eff["mem_style"]
    if "progress_stream" in eff:
        kwargs["callback"] = lambda progress,stream=eff["progress_stream"]: progress | push[stream] | run
    if "progress_callback" in eff:
        kwargs["callback"] = eff["progress_callback"]

    g = load_graph(eff["tag_or_uid"], **kwargs)

    return {"g": g}